from django.utils.decorators import decorator_from_middleware_with_args

from rmr.extensions.middleware.cache import CacheMiddleware


def cache_page(cache_timeout, *, cache=None, key_prefix=None):
    """
    This decorator is very similar to `django.views.decorators.cache.cache_page`
    but instead of `django.middleware.cache.CacheMiddleware` it uses overridden
    version
    """
    return decorator_from_middleware_with_args(CacheMiddleware)(
        cache_timeout=cache_timeout, cache_alias=cache, key_prefix=key_prefix
    )



class DataSet:

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        return '(args={}, kwargs={})'.format(self.args, self.kwargs)


def data_provider(*data_sets):
    def _decorator(test_method):
        test_method.__data_sets__ = data_sets
        return test_method

    return _decorator


class Parametrized(type):

    def __new__(mcs, name, mro, attrs):
        actual_attrs = {}
        for method_name, method in attrs.items():
            data_sets = getattr(method, '__data_sets__', None)
            if not data_sets:
                actual_attrs[method_name] = method
                continue
            for data_set in data_sets:
                test_name = '{}_{}'.format(method_name, data_set)
                actual_attrs[test_name] = (lambda m, ds: (
                    lambda self: m(self, *ds.args, **ds.kwargs)
                ))(method, data_set)
        return super().__new__(mcs, name, mro, actual_attrs)



import logging

from django.core.management.base import BaseCommand as DjangoBaseCommand


class BaseCommand(DjangoBaseCommand):
    """Django command implementation with predefined logging setup

    Use logging.info(), logging.debug() and other logging methods to
    print log messages
    """

    @property
    def logger_name(self):
        return self.__class__.__module__.rsplit('.', 1)[-1]

    @staticmethod
    def get_logger_level(verbosity):
        if verbosity == 1:
            return logging.INFO
        if verbosity == 0:
            return logging.WARNING
        return logging.DEBUG

    def execute(self, *args, **options):
        logging.basicConfig(
            level=self.get_logger_level(options.get('verbosity')),
            format='%(asctime)s [%(levelname)s] ({logger}) %(message)s'.format(
                logger=self.logger_name,
            ),
            datefmt='%Y-%m-%d %H:%M:%S',
        )
        return super().execute(*args, **options)



def dictfetchall(cursor):
    """Return all rows from a cursor as a dict"""
    columns = [col[0] for col in cursor.description]
    return [
        dict(zip(columns, row))
        for row in cursor.fetchall()
    ]



import django.http
import django.test

from django.conf.urls import url
from django.test.utils import override_settings
from django.views.generic import View

import rmr.views

from rmr.utils.test import data_provider, DataSet, Parametrized


class MultipleValuesJsonView(View):

    def post(self, request):
        return django.http.JsonResponse(
            {
                key: request.POST.getlist(key)
                for key in request.POST
            },
            safe=False,
        )


urlpatterns = [
    url(r'simple', lambda request: django.http.JsonResponse(request.POST, safe=False)),
    url(r'multiple', MultipleValuesJsonView.as_view()),
]


@override_settings(ROOT_URLCONF=__name__)
class RequestDecoderTestCase(django.test.TestCase, metaclass=Parametrized):

    def setUp(self):
        self.client = django.test.Client()

    def test_wrong_charset(self):
        response = self.client.post(
            '/simple',
            data=b'\xd0',
            content_type='application/json',
        )
        self.assertIsInstance(response, django.http.HttpResponseBadRequest)
        self.assertEqual(b'bad unicode', response.content)

    @data_provider(
        DataSet(data='"'),
        DataSet(data='{'),
        DataSet(data=']'),
        DataSet(data='hello'),
        DataSet(data='hello=world'),
    )
    def test_wrong_json(self, data):
        response = self.client.post(
            '/simple',
            data=data,
            content_type='application/json',
        )
        self.assertIsInstance(response, django.http.HttpResponseBadRequest)
        self.assertEqual(b'malformed data', response.content)

    @data_provider(
        DataSet(request_data='{"a": 1}', response_data=b'{"a": 1}'),
        DataSet(request_data='{}', response_data=b'{}'),
        DataSet(request_data='""', response_data=b'""'),
        DataSet(request_data='"foo"', response_data=b'"foo"'),
        DataSet(request_data='[]', response_data=b'[]'),
        DataSet(request_data='[1, 2, 3]', response_data=b'[1, 2, 3]'),
        DataSet(request_data='null', response_data=b'null'),
        DataSet(request_data='{"foo": {"bar": "bar"}}',
                response_data=b'{"foo": {"bar": "bar"}}'),
        DataSet(request_data='{"foo": {"bar": [1, 2, 3]}}',
                response_data=b'{"foo": {"bar": [1, 2, 3]}}'),
    )
    def test_valid_json(self, request_data, response_data):
        response = self.client.post(
            '/simple',
            data=request_data,
            content_type='application/json',
        )
        self.assertIsInstance(response, django.http.JsonResponse)
        self.assertEqual(rmr.views.Json.http_code, response.status_code)
        self.assertEqual(response_data, response.content)

    @data_provider(
        DataSet(request_data='{"foo": ["bar", "baz"]}',
                response_data=b'{"foo": ["bar", "baz"]}'),
        DataSet(request_data='{"foo": [{"bar": "bar"}, {"baz": "baz"}]}',
                response_data=b'{"foo": [{"bar": "bar"}, {"baz": "baz"}]}'),
    )
    def test_valid_json_with_multiple_values(self, request_data, response_data):
        response = self.client.post(
            '/multiple',
            data=request_data,
            content_type='application/json',
        )
        self.assertIsInstance(response, django.http.JsonResponse)
        self.assertEqual(rmr.views.Json.http_code, response.status_code)
        self.assertEqual(response_data, response.content)



from datetime import datetime, timedelta

import django.test
from django.utils import timezone

from django.test.utils import override_settings

from rmr.utils.datetime import get_date_range, fromtimestamp
from rmr.utils.test import Parametrized


@override_settings(ROOT_URLCONF=__name__)
class JsonTestCase(django.test.TestCase, metaclass=Parametrized):

    maxDiff = None

    def test_get_date_range_errors(self):
        cases = dict(
            test_exception_with_end_date_is_none_while_max_range_passed=dict(
                start_date=int(datetime(2077, 9, 23).timestamp()),
                end_date=None,
                max_range=86400,
            ),
            test_exception_when_diff_between_start_date_and_end_date_greater_than_max_range=dict(
                start_date=int((datetime.now() - timedelta(seconds=42)).timestamp()),
                end_date=int(datetime.now().timestamp()),
                max_range=1,
            ),
            end_date_not_none_max_range_not_none=dict(
                start_date=None,
                end_date=int(datetime(2241, 7, 25).timestamp()),
                max_range=86400,
                expected_start_date=datetime(2241, 7, 24),
                expected_end_date=datetime(2241, 7, 25),
            ),
        )
        for case, params in cases.items():
            with self.subTest(case=case):
                with self.assertRaises(ValueError):
                    get_date_range(
                        start_date=params['start_date'],
                        end_date=params['end_date'],
                        max_range=params['max_range'],
                    )

    def test_get_date_range_results(self):
        cases = dict(
            all_parameters_none=dict(
                start_date=None,
                end_date=None,
                max_range=None,
                expected_start_date=None,
                expected_end_date=None,
            ),
            start_date_end_date_not_none=dict(
                start_date=int(datetime(2077, 9, 23).timestamp()),
                end_date=None,
                max_range=None,
                expected_start_date=fromtimestamp(datetime(2077, 9, 23).timestamp()),
                expected_end_date=None,
            ),
            max_range_is_none=dict(
                start_date=int(datetime(2077, 9, 23).timestamp()),
                end_date=int(datetime(2241, 7, 25).timestamp()),
                max_range=None,
                expected_start_date=fromtimestamp(datetime(2077, 9, 23).timestamp()),
                expected_end_date=fromtimestamp(datetime(2241, 7, 25).timestamp()),
            ),
            end_date_not_none=dict(
                start_date=None,
                end_date=int(datetime(2241, 7, 25).timestamp()),
                max_range=None,
                expected_start_date=None,
                expected_end_date=fromtimestamp(datetime(2241, 7, 25).timestamp()),
            ),
        )

        for case, params in cases.items():
            with self.subTest(case=case):
                start_date, end_date = get_date_range(
                    start_date=params['start_date'],
                    end_date=params['end_date'],
                    max_range=params['max_range'],
                )
                self.assertEqual(start_date, params['expected_start_date'])
                self.assertEqual(end_date, params['expected_end_date'])



from .hash_lookup import HashLookup
from .pg_trigram import PgTrigramCharField, PgTrigramTextField
from .range import DateRangeField, DateTimeRangeField, FloatRangeField, \
    BigIntegerRangeField, IntegerRangeField
from .pg_ltree import PgLtreeField



import functools

from rmr.errors import ClientError


def authentication_required(fn):
    @functools.wraps(fn)
    def _wrapper(request, *args, **kwargs):
        if not request.user.is_authenticated():
            raise ClientError(
                'Authentication required',
                http_code=401,
                code='authentication_required',
            )
        return fn(request, *args, **kwargs)
    return _wrapper



import contextlib


@contextlib.contextmanager
def patch(obj, attr, value, default=None):
    """
    Context manager inside of which `obj`'s `attr` will be equal `value`
    """
    original = getattr(obj, attr, default)
    setattr(obj, attr, value)
    yield
    setattr(obj, attr, original)



import warnings

from django import forms


class MultipleValueField(forms.TypedMultipleChoiceField):

    def valid_value(self, value):
        return True

    def to_python(self, value):
        return value


class MultiValueField(MultipleValueField):

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "MultiValueField is deprecated use MultipleValueField instead",
            DeprecationWarning,
        )
        super(MultipleValueField, self).__init__(*args, **kwargs)


class BooleanField(forms.Field):

    empty_values = {None}

    def to_python(self, value):
        if value in (True, False):
            return value
        return None


class OffsetLimit(forms.Form):

    offset = forms.IntegerField(required=False, min_value=0)
    limit = forms.IntegerField(required=False)

    @classmethod
    def with_limit_required(cls, *, limit_max_value):
        limit = forms.IntegerField(
            required=True,
            max_value=limit_max_value,
        )
        return type('OffsetLimitRequired', (cls, ), dict(limit=limit))



import contextlib
import functools
import logging

from django.conf import settings
from django.http import JsonResponse, HttpResponse, HttpRequest
from django.utils.decorators import classonlymethod
from django.views.decorators.http import last_modified
from django.utils.functional import lazy
from django.views.generic import View

import rmr

from rmr.utils.cache import cache_page
from rmr.utils.decorators import conditional

request_logger = logging.getLogger('rmr.request')

response_logger = logging.getLogger('rmr.response')


class Json(View):

    http_code = 200

    logger = logging.getLogger('rmr.view')

    def __init__(self, request: HttpRequest=None, **kwargs):
        super().__init__(**kwargs)
        self.request = request

    @classmethod
    def expires(cls):
        """
        Lazy evaluated 'max-age' value of Cache-Control header
        """
        return settings.CACHE_MIDDLEWARE_SECONDS

    @classmethod
    def last_modified(cls, request, *args, **kwargs):
        """
        Lazy evaluated value of Last-Modified header
        """
        pass

    @classmethod
    def cache_headers_allowed(cls, request, *args, **kwargs):
        return request.method in ('GET', 'HEAD')

    @classonlymethod
    def as_view(cls, **initkwargs):
        view = patched_view = super().as_view(**initkwargs)
        last_modified_evaluator = functools.lru_cache()(cls.last_modified)
        patched_view = last_modified(last_modified_evaluator)(patched_view)
        patched_view = cache_page(
            lazy(cls.expires, int)(),
            key_prefix=last_modified_evaluator,
        )(patched_view)
        view = conditional(cls.cache_headers_allowed, patched_view)(view)

        @functools.wraps(cls.as_view)
        def logging_view(request, *args, **kwargs):

            request_logger.debug(
                'request_method: %(request_method)s, '
                'request_path: %(request_path)s, '
                'request_headers: %(request_headers)s, '
                'request_params: %(request_params)s, '
                'request_data: %(request_data)s, ',
                dict(
                    request_method=request.method,
                    request_path=request.path,
                    request_headers=request.META,
                    request_params=request.GET,
                    request_data=request.POST,
                ),
            )

            response = view(request, *args, **kwargs)

            response_logger.debug(
                'response_code: %(response_code)s, '
                'response_headers: %(response_headers)s, '
                'response_data: %(response_data)s',
                dict(
                    response_code=response.status_code,
                    response_headers=response.items(),
                    response_data=response.content,
                ),
            )

            return response

        return logging_view

    def dispatch(self, request: HttpRequest, *args, **kwargs):
        self.request = request
        try:
            result = super().dispatch(request, *args, **kwargs)
            if isinstance(result, HttpResponse):
                return result
            http_code = self.http_code
            api_result = dict(
                data=result,
            )
        except rmr.Error as error:

            response_logger.log(
                error.level,
                '%(code)s: %(message)s',
                dict(message=error.message, code=error.code),
            )

            http_code = error.http_code
            api_result = dict(
                error=dict(
                    code=error.code,
                    description=error.message,
                ),
            )

        response = JsonResponse(api_result, status=http_code, safe=False)
        response.setdefault('Content-Length', len(response.content))
        return response

    @staticmethod
    def get_range(offset=None, limit=None, limit_default=None, limit_max=None):
        # TODO make another method with offset/limit validation delegated to rmr.forms.OffsetLimit
        # TODO make deprecated
        start = 0
        stop = None
        with contextlib.suppress(ValueError):
            start = offset and int(offset) or start
            if start < 0:
                raise rmr.ClientError(
                    'Offset must be a positive number',
                    code='incorrect_offset',
                )
            limit = limit and int(limit) or limit_default
            if limit is not None:
                if limit <= 0:
                    raise rmr.ClientError(
                        'Limit must be a positive number',
                        code='incorrect_limit',
                    )
                if limit_max and limit > limit_max:
                    raise rmr.ClientError(
                        'Maximum value of limit must be '
                        'less then or equal {}'.format(limit_max),
                        code='max_limit_exceeded',
                    )
                stop = start + limit
            elif limit_max:
                raise rmr.ClientError(
                    'Limit must be provided',
                    code='limit_not_provided',
                )

            return start, stop

        raise rmr.ClientError(
            'Limit and offset must be a numbers if provided',
            code='incorrect_limit_or_offset',
        )



import json

from django.conf import settings
from django.http import HttpResponseBadRequest, QueryDict


class RequestDecoder:

    content_type = 'application/json'

    allowed_methods = {
        'POST', 'PUT', 'PATCH',
    }

    def process_request(self, request):
        if request.method not in self.allowed_methods:
            return
        content_type = request.META.get('CONTENT_TYPE', '')
        if not content_type.startswith(self.content_type):
            return
        encoding = request.encoding or settings.DEFAULT_CHARSET
        try:
            body = request.body.decode(encoding=encoding)
        except UnicodeDecodeError:
            return HttpResponseBadRequest('bad unicode')
        try:
            request.POST = self.json_decode(body, encoding=encoding)
        except ValueError:
            return HttpResponseBadRequest('malformed data')

    @staticmethod
    def json_decode(body, encoding=None):
        data = json.loads(body, encoding=encoding)
        if not isinstance(data, dict):
            # all data of type other then dict will be returned as is
            return data
        query_dict = QueryDict(mutable=True)
        for key, value in data.items():
            if isinstance(value, list):
                # treat list values as set of values of MultiValueDict
                query_dict.setlist(key, value)
            else:
                query_dict[key] = value
        query_dict._mutable = False
        return query_dict



from django.db import models

from rmr.utils.hash import crc64


class HashLookup(models.BigIntegerField):

    def __init__(self, *args, object_field, **kwargs):
        self.object_field = object_field
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs['object_field'] = self.object_field
        return name, path, args, kwargs

    def pre_save(self, model_instance, add):
        obj = getattr(model_instance, self.object_field)
        return None if obj is None else crc64(obj)

    def get_prep_lookup(self, lookup_type, value):
        return crc64(value)



from django.db import models
from django.db.models import Lookup


class PgLtreeField(models.CharField):
    description = 'ltree (up to %(max_length)s)'

    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = 255
        super().__init__(*args, **kwargs)

    def db_type(self, connection):
        return 'ltree'

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs['max_length']
        return name, path, args, kwargs


@PgLtreeField.register_lookup
class AncestorOf(Lookup):
    lookup_name = 'ancestor_of'

    def as_sql(self, qn, connection):
        lhs, lhs_params = self.process_lhs(qn, connection)
        rhs, rhs_params = self.process_rhs(qn, connection)
        params = lhs_params + rhs_params
        return '%s @> %s' % (lhs, rhs), params


@PgLtreeField.register_lookup
class DescendantOf(Lookup):
    lookup_name = 'descendant_of'

    def as_sql(self, qn, connection):
        lhs, lhs_params = self.process_lhs(qn, connection)
        rhs, rhs_params = self.process_rhs(qn, connection)
        params = lhs_params + rhs_params
        return '%s <@ %s' % (lhs, rhs), params


@PgLtreeField.register_lookup
class Match(Lookup):
    lookup_name = 'match'

    def as_sql(self, qn, connection):
        lhs, lhs_params = self.process_lhs(qn, connection)
        rhs, rhs_params = self.process_rhs(qn, connection)
        params = lhs_params + rhs_params
        return '%s ~ %s' % (lhs, rhs), params
